/*
	PROforma, PROGS Font & Raster Manager

	routines to control the QVME screen driver

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 28 September 1995
	02.11.1995 made alloc/realease window much safer
	14.12.1995 automatic recognition of mono/mode4
	21.06.1996 support for setting DisplayMode added
*/

#include "mem_h"
#include "job_h"
#include "win_h"

#include "gstate_h"

static short screenXsize;	/* total screen size - horizontal */
static short screenYsize;	/* total screen size - vertical */
static WindowMode mode; 	/* screen mode */

extern DRIVdriver init; 	/* the actual driver description */

Error ColourMonoAdjust(Gstate gstate, pt *colour);
Error ColourMonoNearest(Gstate gstate, pt *colour, int *result);
Error ColourMonoDelta(Gstate gstate, pt *colour, pt *delta);

typedef struct _DriverInfo {
	Window window;			/* the window which displays the output */
	WindowArea *area;		/* the bitmap which will be displayed in it */
	WindowInfo info;		/* the size and position of the window */
	WindowInfo rest;		/* the position in the window to display the area */
	WindowArea *save;		/* saved buffer */
	Job update; 			/* job id of update job if exists */
	} DriverInfo;

Error scrsize()
{
	Error err;
	WindowInfo limits;
	Window win;

	err=WINOpen(2,1,0,0,&win);
	if (err) return err;
	err=WINLimits(win,&limits);
	if (err) return err;
	screenXsize=limits.xsiz;
	screenYsize=limits.ysiz;
	WINClose(win);
	return err;
}

Error InitDriver()
{
	Error err;
	Job job;

	/* need to determine the size of the window - safely */
	err=JOBCall(scrsize,"determine screen size",8,1024,&job,0);
	if (err) return err;
	JOBWait(job,FOREVER);

	/* read screen mode and adjust bitmap when mono */
	WINMode(&mode);
	if (mode==WINDOWMODE_2)
	{
		init.bitmap.PlaneCount=1;
		init.bitmap.PlaneGrouping=BITMAP_NO_PLANES;
		init.ColourSpace=PF_COLOURSPACE_GRAYSHADE;
		init.colour.Gray.ColourAdjust =ColourMonoAdjust;
		init.colour.Gray.ColourNearest=ColourMonoNearest;
		init.colour.Gray.ColourDelta  =ColourMonoDelta;
	}

	return err;
}

Error WindowMove(Gstate gstate, int dx, int dy)
{
	DriverInfo *di=gstate->PrinterExtra;

	di->info.xpos+=dx;
	di->info.ypos+=dy;
	di->rest.xpos=di->info.xpos&1;

	WINRedefine(di->window,0,0,
				di->info.xsiz+di->rest.xpos*2+(di->info.xsiz&1), di->info.ysiz,
				di->info.xpos&-2,di->info.ypos);
	return ERR_OK;
}

Error BufferSize(Gstate gstate, int xsiz, int ysiz, Size *size)
{
	Error err;
	err=gstate->bitmap->BufferSize(xsiz,ysiz,size);
	*size+=offsetof(WindowArea,data);
	return err;
}

Error BufferUpdate(Gstate gstate)
{
	DriverInfo *di=gstate->PrinterExtra;

	return WINAreaSave(di->window, &di->rest, 0,0, di->area);
}

Error BufferSave(Gstate gstate)
{
	DriverInfo *di=gstate->PrinterExtra;
	Size size=di->area->line_increment*di->area->ysiz+offsetof(WindowArea,data);

	if (!di->save)
		if (MEMAllocate(size,&di->save)) return ERR_IMEM;
	MEMCopy(size,di->area,di->save);
	return ERR_OK;
}

Error BufferRestore(Gstate gstate)
{
	DriverInfo *di=gstate->PrinterExtra;
	Size size=di->area->line_increment*di->area->ysiz+offsetof(WindowArea,data);

	if (!di->save) return ERR_ISYN;
	MEMCopy(size,di->save,di->area);
	return ERR_OK;
}

Error InitInit(Gstate gstate, char *device, PBbox *area)
{
	Error err;
	DriverInfo *di;

	err=MEMAllocate(sizeof(DriverInfo),&di);
	if (err) return err;

	err=WINOpen(2,1,0,0,&di->window);
	if (err) { MEMRelease(di); return err; };
	err=WINLimits(di->window,&di->info);
	if (err) { MEMRelease(di); return err; };

	gstate->PrinterExtra=di;
	gstate->todev.x = pt_one * screenXsize / 720;
	gstate->todev.y = pt_one * screenYsize / 540;
	gstate->ppp.x = pt_one * 720 / screenXsize;
	gstate->ppp.y = pt_one * 540 / screenYsize;

	MEMCopy(sizeof(WindowInfo),&di->info,area);

	return err;
}

Error InitDone(Gstate gstate)
{
	DriverInfo *di=gstate->PrinterExtra;

	/* have to build the header of the WindowArea */
	di->area=(WindowArea *)gstate->raster->base;
	gstate->raster->base+=offsetof(WindowArea,data);
	di->area->flag=WINDOWAREA_FLAG;
	di->area->xsiz=gstate->raster->all.xsiz;
	di->area->ysiz=gstate->raster->all.ysiz;
	di->area->line_increment=gstate->raster->linc;
	di->area->mode=mode;

	/* fill in the info stuff in the DriverInfo structure */
	di->info.xsiz=di->rest.xsiz=gstate->pagebbox.xsiz;
	di->info.ysiz=di->rest.ysiz=gstate->pagebbox.ysiz;
	di->info.xpos=				pt2short(gstate->pagebbox.xporg);
	di->info.ypos=				pt2short(gstate->pagebbox.yporg);

	/* and set the window */
	return WindowMove(gstate,0,0);
}

Error CleanUp(Gstate gstate)
{
	DriverInfo *di=gstate->PrinterExtra;

	if (di)
	{
		if (di->save) MEMRelease(di->save);
		WINClose(di->window);
		MEMRelease(gstate->PrinterExtra);
	}
	return ERR_OK;
}

Error ShowPage(Gstate gstate)
{
	DriverInfo *di=gstate->PrinterExtra;

	if (di->update)
	{
		JOBZap(di->update,ERR_OK);
		di->update=NUL;
	}
	return WINAreaRestore(di->window, &di->rest, 0,0, di->area);
}

static update(DriverInfo *di)
{
	while (TRUE)
	{
		JOBSuspend(MYSELF,5,NULL);
		WINAreaRestore(di->window, &di->rest, 0,0, di->area);
	}
}

Error DisplayMode(Gstate gstate, int mode)
{
	DriverInfo *di=gstate->PrinterExtra;

	switch(mode)
	{
	case PF_DISPLAYMODE_NORMAL:
		if (di->update) JOBZap(di->update,ERR_OK);
		di->update=NUL;
		break;
	case PF_DISPLAYMODE_UPDATE:
		if (!di->update)
			return JOBCall(update,"display screen",1,2048,&di->update,1,di);
		break;
	default:
		return ERR_IPAR;
	}
	return ERR_OK;
}

Error ColourNearest(Gstate gstate, ColourRGB *colour, int *result)
{
	*result=0;						/* assume black */
	if (colour->green>pt_hundred/2)
		*result+=2; 				/* add some green */
	if (colour->red  >pt_hundred/2)
		*result+=1; 				/* add some red */
	return ERR_OK;
}

Error ColourDelta(Gstate gstate, ColourRGB *colour, ColourRGB *delta)
{
	MEMCopy(sizeof(ColourRGB),colour,delta);
	delta->blue=colour->blue;
	if (colour->green>pt_hundred/2)
		delta->green-=pt_hundred;
	if (colour->red  >pt_hundred/2)
		delta->red	-=pt_hundred;
	return ERR_OK;
}

Error ColourMonoAdjust(Gstate gstate, pt *colour)
{
	if (*colour>=pt_hundred)
		*colour=0;
	else
	if (*colour<=0)
		*colour=pt_hundred;
	else
		*colour=pt_hundred-*colour;
	return ERR_OK;
}

Error ColourMonoNearest(Gstate gstate, pt *colour, int *result)
{
	*result=0;						/* assume white */
	if (*colour>pt_hundred/2)
		*result=1;					/* make it black */
	return ERR_OK;
}

Error ColourMonoDelta(Gstate gstate, pt *colour, pt *delta)
{
	*delta=*colour;
	if (*colour>pt_hundred/2)
		*delta-=pt_hundred;
	return ERR_OK;
}
