/*
	PROforma, PROGS Font & Raster Manager

	Canon BJ-10 bubblejet printer driver

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 1 May 1996, based on Epson 24 pin printer driver and BJ-10
						printer driver for PROforma v1.06
	27.06.1996 should now work on all Canon BJ printers
*/

#include "io_h"
#include "mem_h"
#include "gstate_h"

typedef struct _DriverInfo {
	Channel channel;		/* the output channel for this driver */
	int bit;				/* next bit for transform */
	int xskip, yskip;		/* skip values to end up at the topleft */
	char *buffer;
} DriverInfo;


Bbox PrintableArea={
	double2pt(577.8), short2pt(780),/* size */
	double2pt(8.6), short2pt(24)	/* origin */
};

char PrintDevice[IO_MAXFULLNAME]="pard";

Error Define(char *name, char *value)
{
	double x,y;

	if (STRSameCD(name,"DEFAULT-DEVICE"))
	{
		STRCopySafe(value,PrintDevice,IO_MAXFULLNAME);
		return ERR_OK;
	}
	if (STRSameCD(name,"PRINTABLE-AREA-SIZE"))
	{
		STRFormatGet(value,NULL,"%f %f",&x,&y);
		PrintableArea.xsiz=double2pt(x);
		PrintableArea.ysiz=double2pt(y);
		return ERR_OK;
	}
	if (STRSameCD(name,"PRINTABLE-AREA-ORIGIN"))
	{
		STRFormatGet(value,NULL,"%f %f",&x,&y);
		PrintableArea.xorg=double2pt(x);
		PrintableArea.yorg=double2pt(y);
		return ERR_OK;
	}
	return ERR_IPAR;
}

Error InitInit(Gstate gstate, char *device, PBbox *area)
{
	Error err;
	DriverInfo *di;
	DRIVdriver *prt=gstate->printer;

	err=MEMAllocate(sizeof(DriverInfo),&di);
	if (err) return err;

	if (device && *device)
		err=IOOpen(device,OPEN_OVER,&di->channel);
	else
		err=IOOpen(PrintDevice,OPEN_OVER,&di->channel);
	if (err) { MEMRelease(di); return err; };

	gstate->PrinterExtra=di;
	gstate->todev.x = pt_one * prt->xdpi / 72;
	gstate->todev.y = pt_one * prt->ydpi / 72;
	gstate->ppp.x = pt_one * 72 / prt->xdpi;
	gstate->ppp.y = pt_one * 72 / prt->ydpi;

	area->xsiz=pt2rshort(fixmul(PrintableArea.xsiz,gstate->todev.x));
	area->ysiz=pt2rshort(fixmul(PrintableArea.ysiz,gstate->todev.y));
	area->xorg=pt2rshort(fixmul(PrintableArea.xorg,gstate->todev.x));
	area->yorg=pt2rshort(fixmul(PrintableArea.yorg,gstate->todev.y));

	return err;
}

Error InitDone(Gstate gstate)
{
	DriverInfo *di=gstate->PrinterExtra;
	Size linc;

	gstate->bitmap->BufferSize(gstate->pagebbox.xsiz,48,&linc);

	/* complete the initialisation of the DriverInfo */
	di->buffer=(char *)gstate->raster->base;
	gstate->raster->base+=linc;

	di->xskip=pt2rshort(gstate->pagebbox.xporg-fixmul(PrintableArea.xorg,
							   gstate->todev.x));
	di->yskip=pt2rshort(gstate->pagebbox.yporg-fixmul(PrintableArea.yorg,
							   gstate->todev.y));

	return ERR_OK;
}

Error BufferSize(Gstate gstate, int xsiz, int ysiz, Size *size)
{
	DriverInfo *di=gstate->PrinterExtra;
	return gstate->bitmap->BufferSize(xsiz,ysiz+48,size);
}

Error CleanUp(Gstate gstate)
{
	DriverInfo *di=gstate->PrinterExtra;

	if (di)
	{
		IOClose(di->channel);
		MEMRelease(gstate->PrinterExtra);
	}
	return ERR_OK;
}

Error ColourNearest(Gstate gstate, pt *colour, int *result)
{
	*result=0;						/* assume white */
	if (*colour>pt_hundred/2)
		*result=1;					/* make it black */
	return ERR_OK;
}

Error ColourDelta(Gstate gstate, pt *colour, pt *delta)
{
	*delta=*colour;
	if (*colour>pt_hundred/2)
		*delta-=pt_hundred;
	return ERR_OK;
}

/* !!!!!!!!!!! code to send data to printer !!!!!!!!!!! */

/* some macros for generating output */
#define printb(s,b) IOPutRawBytes(chid,s,b,NULL)
#define prints(s)	printb(strlen(s),s)
#define printr(s)	printb(sizeof(s)-1,s);
#define printc(c)	{ char s=c; IOPutRawBytes(chid,1,&s,NULL); }

/* routine to fill in a scan line in the transposed data */
static const char setbit[8] = { 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01};
static void transpose(char *data, char *dest, int bit, int skip, int linelen)
{
	char mask=setbit[bit&7];
	int i,j;

	dest+=bit>>3;
	for (j=linelen; j; j--)
	{
		char test=*data++;
		for (i=0; i<8; i++)
		{
			if (test & setbit[i]) *dest|=mask;
			dest+=skip;
		}
	}
}

/* routine to send a line to the printer */
static void printline(Gstate gstate, char *buff, int count)
{
	DriverInfo *di=gstate->PrinterExtra;
	Channel chid=di->channel;
	int xskip=di->xskip/3;
	int xcount=count/6;

	/* make sure x position is correct */
	if (xskip && xcount)
	{
		prints("\033d");
		printc(xskip & 0xff);
		printc(xskip >> 8);
	}

	/* print the actual data */
	if (xcount)
	{
		prints("\033CB\x04\x04");
		printc(xcount & 0xff);
		printc(xcount >> 8);
		printb(count,buff);
	}
	prints("\r");
}


Error InitPage(Gstate gstate)
{
	DriverInfo *di=gstate->PrinterExtra;
	Channel chid=di->channel;
	int yskip=di->yskip;

	/* set graphics line spacing unit and auto line feed */
	printr("\x1b[\0x04\0\0\0\x68\x01\0335\x01");
	if (yskip)
	{
		printr("\0333\xff");
		while(yskip>255)
		{
			prints("\r");
			yskip-=255;
		}
		prints("\0333");
		printc(yskip);
		prints("\r");
	}

	/* Initialize the line skip */
	printr("\0333\x30");
	di->bit=0;
	return ERR_OK;
}

Error ShowPage(Gstate gstate)
{
	DriverInfo *di=gstate->PrinterExtra;
	Channel chid=di->channel;
	int line_size=gstate->raster->linc;
	char *data=(char *)gstate->raster->base;
	char *buff=di->buffer;
	char *buff_end; int i;

	for(i=gstate->pagebbox.ysiz ; i ; i--, data+=line_size)
	{
		/* clear the buffer if necessary */
		if (di->bit==0) MEMClear(line_size*48,di->buffer);

		/* Add this scan line in the buffer (byteline) */
		transpose(data,buff,di->bit,6,line_size);
		di->bit++;

		if (di->bit==48)
		{
			buff_end=buff+line_size*48;
			/* remove trailing zeros */
			while (buff_end>buff && !buff_end[-1]) buff_end--;

			printline(gstate, buff, buff_end-buff+6-1);
			di->bit=0;
		}
	}
	return ERR_OK;
}

Error StopPage(Gstate gstate)
{
	DriverInfo *di=gstate->PrinterExtra;
	Channel chid=di->channel;
	int line_size = gstate->raster->linc;
	char *data=(char *)gstate->raster->base;
	char *buff=di->buffer;
	char *buff_end=buff+line_size*48;

	if (di->bit)
	{
		/* remove trailing zeros */
		while (buff_end>buff && !buff_end[-1]) buff_end--;
		printline(gstate, buff, buff_end-buff+6-1);
	}
	prints("\f");
	return ERR_OK;
}


/* !!!!!!!!!!! printer driver definition !!!!!!!!!!! */

static Error nimp() 	{ return ERR_NIMP; }
static Error nothing()	{ return ERR_OK; }

Error InitDriver();

Error SetCopies(Gstate, int count);
Error SetDisplayMode(Gstate, int mode);
Error WindowMove(Gstate, int dx, int dy);
Error BufferUpdate(Gstate);
Error ColourAdjust(Gstate, pt *);
Error GetFullPageBbox(Gstate, Bbox *box);
Error DefaultDevice(Gstate, char *device);

#define InitDriver		nothing
#define WindowMove		nothing
#define ColourAdjust	nothing
#define SetDisplayMode	nimp
#define GetFullPageBbox nimp
#define DefaultDevice	nimp
#define SetCopies		nimp
#define BufferUpdate	(Error (*)(Gstate))nimp
#define BufferSave		(Error (*)(Gstate))nimp
#define BufferRestore	(Error (*)(Gstate))nimp
#define Handle			(Error (*)(int,...))nimp

DRIVdriver init = {
	NULL, PF_PRINTERDRIVER,
	"Canon BJ-10, 360dpi",		/* name */
	360, 360,					/* xdpi, ydpi */
	{                           /* mono bitmap */
		1, 1,					/* 1 plane, 1 pixel deep */
		BITMAP_NO_BYTEGROUPING, /* no byte grouping */
		BITMAP_NO_PLANES,		/* no planes */
	},

	InitDriver,
	Define,

	InitInit,
	BufferSize,
	InitDone,
	CleanUp,

	InitPage,
	ShowPage,
	StopPage,

	SetCopies,
	SetDisplayMode,
	WindowMove,

	BufferUpdate,
	BufferSave,
	BufferRestore,

	PF_COLOURSPACE_GRAYSHADE,
	{
		ColourAdjust,
		ColourNearest,
		ColourDelta
	},

	GetFullPageBbox,
	DefaultDevice,

	Handle
};
